/*
	NO LICENSE
*/
#include <ilbc/iLBC_encode.h>
#include <ilbc/iLBC_decode.h>

#include "msfilter.h"

typedef struct EncState{
	int nsamples;
	int nbytes;
	int ms_per_frame;
	uint32_t ts;
	MSBufferizer bufferizer;
	iLBC_Enc_Inst_t ilbc_enc;	
}EncState;

static void enc_init(MSFilter *f){
	EncState *s=ms_new(EncState,1);
	s->nsamples=BLOCKL_20MS;
	s->nbytes=NO_OF_BYTES_20MS;
	s->ms_per_frame=20;
	s->ts=0;
	ms_bufferizer_init(&s->bufferizer);
	f->data=s;
}

static void enc_uninit(MSFilter *f){
	EncState *s=(EncState*)f->data;
	ms_bufferizer_uninit(&s->bufferizer);
	ms_free(f->data);
}

static void enc_preprocess(MSFilter *f){
	EncState *s=(EncState*)f->data;
	initEncode(&s->ilbc_enc,s->ms_per_frame);
}

static int enc_set_fmtp(MSFilter *f, void *arg){
	const char *fmtp=(const char *)arg;
	EncState *s=(EncState*)f->data;
	if (strstr(fmtp,"ptime=20")!=NULL){
		s->nsamples=BLOCKL_20MS;
		s->nbytes=NO_OF_BYTES_20MS;
		s->ms_per_frame=20;
	}else if (strstr(fmtp,"ptime=30")!=NULL){
		s->nsamples=BLOCKL_30MS;
		s->nbytes=NO_OF_BYTES_30MS;
		s->ms_per_frame=30;
	}
	return 0;
}

static void enc_process(MSFilter *f){
	EncState *s=(EncState*)f->data;
	mblk_t *im,*om;
	int size=s->nsamples*2;
	int16_t samples[BLOCKL_MAX];/* the largest size */
	float samples2[BLOCKL_MAX];
	int i;
	
	while((im=ms_queue_get(f->inputs[0]))!=NULL){
		ms_bufferizer_put(&s->bufferizer,im);
	}
	while(ms_bufferizer_read(&s->bufferizer,(uint8_t*)samples,size)==size){
		om=allocb(s->nbytes,0);
		for (i=0;i<s->nsamples;i++){
			samples2[i]=samples[i];
		}
		iLBC_encode((uint8_t*)om->b_wptr,samples2,&s->ilbc_enc);
		om->b_wptr+=s->nbytes;
		mblk_set_timestamp_info(om,s->ts);
		s->ts+=s->nsamples;
		ms_queue_put(f->outputs[0],om);
	}
}

static MSFilterMethod enc_methods[]={
	{	MS_FILTER_SET_FMTP,		enc_set_fmtp },
	{	0								,		NULL			}
};

MSFilterDesc ms_ilbc_enc_desc={
	.id=MS_FILTER_PLUGIN_ID,
	.name="MSIlbcEnc",
	.text="iLBC encoder",
	.category=MS_FILTER_ENCODER,
	.enc_fmt="iLBC",
	.ninputs=1,
	.noutputs=1,
	.init=enc_init,
	.preprocess=enc_preprocess,
	.process=enc_process,
	.uninit=enc_uninit,
	.methods=enc_methods
};

typedef struct DecState{
	int nsamples;
	int nbytes;
	int ms_per_frame;
	iLBC_Dec_Inst_t ilbc_dec;	
}DecState;


static void dec_init(MSFilter *f){
	DecState *s=ms_new(DecState,1);
	s->nsamples=0;
	s->nbytes=0;
	s->ms_per_frame=0;
	f->data=s;
}

static void dec_uninit(MSFilter *f){
	ms_free(f->data);
}

static void dec_process(MSFilter *f){
	DecState *s=(DecState*)f->data;
	mblk_t *im,*om;
	int nbytes;
	float samples[BLOCKL_MAX];
	int i;
	while ((im=ms_queue_get(f->inputs[0]))!=NULL){
		nbytes=msgdsize(im);
		if (s->nbytes==0){
			/* must init the decoder first */
			if (nbytes<NO_OF_BYTES_30MS){
				/* we assume we are in the 20 ms mode */
				s->ms_per_frame=20;
				s->nbytes=NO_OF_BYTES_20MS;
				s->nsamples=BLOCKL_20MS;
			}else{
				s->ms_per_frame=30;
				s->nbytes=NO_OF_BYTES_30MS;
				s->nsamples=BLOCKL_30MS;
			}
			initDecode(&s->ilbc_dec,s->ms_per_frame,0);
		}
		if (nbytes>=s->nbytes){
			om=allocb(s->nsamples*2,0);
			iLBC_decode(samples,(uint8_t*)im->b_rptr,&s->ilbc_dec,1);
			for (i=0;i<s->nsamples;i++,om->b_wptr+=2){
				*((int16_t*)om->b_wptr)=samples[i];
			}
			ms_queue_put(f->outputs[0],om);
		}else{
			ms_warning("bad iLBC frame !");
		}
		freemsg(im);
	}
}

MSFilterDesc ms_ilbc_dec_desc={
	.id=MS_FILTER_PLUGIN_ID,
	.name="MSIlbcDec",
	.text="iLBC decoder",
	.category=MS_FILTER_DECODER,
	.enc_fmt="iLBC",
	.ninputs=1,
	.noutputs=1,
	.init=dec_init,
	.process=dec_process,
	.uninit=dec_uninit
};

void libmsilbc_init(){
	ms_filter_register(&ms_ilbc_enc_desc);
	ms_filter_register(&ms_ilbc_dec_desc);
}
